//! Node filters.

use std::collections::HashSet;
use std::convert::AsRef;
use std::iter::Iterator;
use std::str::FromStr;

use clap::Args;
use glob::Pattern as GlobPattern;
use serde::Deserialize;

use super::{ColmenaError, ColmenaResult, NodeConfig, NodeName};

#[derive(Debug, Default, Args)]
pub struct NodeFilterOpts {
    #[arg(
        long,
        value_name = "NODES",
        help = "Node selector",
        long_help = r#"Select a list of nodes to deploy to.

The list is comma-separated and globs are supported. To match tags, prepend the filter by @. Valid examples:

- host1,host2,host3
- edge-*
- edge-*,core-*
- @a-tag,@tags-can-have-*"#
    )]
    pub on: Option<NodeFilter>,
}

/// Which fields need to be evaluated
/// in order to execute the node filter.
#[derive(Clone, Copy, Debug, Default)]
pub struct NeedsEval {
    /// Need to evaluate deployment.aliases of all nodes.
    pub aliases: bool,
    /// Need to evaluate deployment.tags of all nodes.
    pub tags: bool,
}

impl NeedsEval {
    pub fn any(&self) -> bool {
        self.aliases || self.tags
    }
}

impl std::ops::BitOr for NeedsEval {
    type Output = Self;
    fn bitor(self, rhs: Self) -> Self::Output {
        Self {
            aliases: self.aliases || rhs.aliases,
            tags: self.tags || rhs.tags,
        }
    }
}

impl std::ops::BitOrAssign for NeedsEval {
    fn bitor_assign(&mut self, rhs: Self) {
        *self = *self | rhs;
    }
}

#[derive(Debug, Default, Deserialize)]
pub struct PartialNodeConfig {
    pub aliases: Option<Vec<NodeName>>,
    pub tags: Option<Vec<String>>,
}

impl From<NodeConfig> for PartialNodeConfig {
    fn from(node_config: NodeConfig) -> Self {
        Self {
            aliases: Some(node_config.aliases),
            tags: Some(node_config.tags),
        }
    }
}

/// A filter rule.
#[derive(Debug, Clone, Eq, PartialEq)]
pub enum NodeFilter {
    /// Matches a node's attribute name.
    MatchName(GlobPattern),

    /// Matches a node's `deployment.tags`.
    MatchTag(GlobPattern),

    /// Matches an Union
    Union(Vec<Box<NodeFilter>>),

    /// Matches an Intersection
    Inter(Vec<Box<NodeFilter>>),

    /// Matches the complementary
    Not(Box<NodeFilter>),

    /// Empty
    Empty,
}

impl FromStr for NodeFilter {
    type Err = ColmenaError;
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Self::new(s)
    }
}

#[inline]
fn end_delimiter(c: char) -> bool {
    [',', '&', ')'].contains(&c)
}

impl NodeFilter {
    fn and(a: Self, b: Self) -> Self {
        match (a, b) {
            (Self::Inter(mut av), Self::Inter(mut bv)) => {
                av.append(&mut bv);
                Self::Inter(av)
            }
            (Self::Inter(mut av), b) => {
                av.push(Box::new(b));
                Self::Inter(av)
            }
            (a, Self::Inter(mut bv)) => {
                bv.push(Box::new(a));
                Self::Inter(bv)
            }
            (a, b) => Self::Inter(vec![Box::new(a), Box::new(b)]),
        }
    }

    fn or(a: Self, b: Self) -> Self {
        match (a, b) {
            (Self::Union(mut av), Self::Union(mut bv)) => {
                av.append(&mut bv);
                Self::Union(av)
            }
            (Self::Union(mut av), b) => {
                av.push(Box::new(b));
                Self::Union(av)
            }
            (a, Self::Union(mut bv)) => {
                bv.push(Box::new(a));
                Self::Union(bv)
            }
            (a, b) => Self::Union(vec![Box::new(a), Box::new(b)]),
        }
    }

    fn not(a: Self) -> Self {
        if let Self::Not(ae) = a {
            *ae
        } else {
            Self::Not(Box::new(a))
        }
    }

    /// Parses an elementary expression,
    /// that is base tags and name, with expression between parentheses
    /// Negations are also parsed here as the most prioritary operation
    ///
    /// It returns the unparsed text that follows
    fn parse_expr0(unparsed: &str) -> ColmenaResult<(Self, &str)> {
        let unparsed = unparsed.trim_start();
        // Negation
        if let Some(negated_expr) = unparsed.strip_prefix('!') {
            let (negated, unparsed) = Self::parse_expr0(negated_expr)?;
            Ok((Self::not(negated), unparsed))
        } else
        // parentheses
        if let Some(parenthesed_expr) = unparsed.strip_prefix('(') {
            let (interior, unparsed) = Self::parse_expr2(parenthesed_expr)?;
            Ok((
                interior,
                unparsed.strip_prefix(')').ok_or(ColmenaError::Unknown {
                    message: format!("Expected a closing parenthesis at {:?}.", unparsed),
                })?,
            ))
        } else
        // tag
        if let Some(tag_expr) = unparsed.strip_prefix('@') {
            match tag_expr
                .find(end_delimiter)
                .map(|idx| tag_expr.split_at(idx))
                .map(|(tag, end)| (tag.trim_end(), end))
            {
                Some((tag, unparsed)) => {
                    if tag.is_empty() {
                        return Err(ColmenaError::EmptyFilterRule);
                    } else {
                        Ok((Self::MatchTag(GlobPattern::new(tag).unwrap()), unparsed))
                    }
                }
                None => {
                    let tag_expr = tag_expr.trim_end();
                    if tag_expr.is_empty() {
                        Err(ColmenaError::EmptyFilterRule)
                    } else {
                        Ok((Self::MatchTag(GlobPattern::new(tag_expr).unwrap()), ""))
                    }
                }
            }
        } else
        //node name
        {
            match unparsed
                .find(end_delimiter)
                .map(|idx| unparsed.split_at(idx))
                .map(|(tag, end)| (tag.trim_end(), end))
            {
                Some((name, unparsed)) => {
                    if name.is_empty() {
                        Err(ColmenaError::EmptyFilterRule)
                    } else {
                        Ok((Self::MatchName(GlobPattern::new(name).unwrap()), unparsed))
                    }
                }
                None => {
                    let unparsed = unparsed.trim_end();
                    if unparsed.is_empty() {
                        Err(ColmenaError::EmptyFilterRule)
                    } else {
                        Ok((Self::MatchName(GlobPattern::new(unparsed).unwrap()), ""))
                    }
                }
            }
        }
    }

    /// Parses the union operations between elementary expression.
    ///
    /// It returns the unparsed text that follows
    fn parse_op1(acc: Self, unparsed: &str) -> ColmenaResult<(Self, &str)> {
        let unparsed = unparsed.trim_start();
        if let Some(unions) = unparsed.strip_prefix(',') {
            let (base_expr, unparsed) = Self::parse_expr0(unions)?;
            Self::parse_op1(Self::or(acc, base_expr), unparsed)
        } else {
            Ok((acc, unparsed))
        }
    }

    /// Parses elementary expression and their unions.
    ///
    /// It returns the unparsed text that follows
    fn parse_expr1(unparsed: &str) -> ColmenaResult<(Self, &str)> {
        let (base_expr, unparsed) = Self::parse_expr0(unparsed)?;
        Self::parse_op1(base_expr, unparsed)
    }

    /// Parses the intersection operations between unions.
    ///
    /// It returns the unparsed text that follows
    fn parse_op2(acc: Self, unparsed: &str) -> ColmenaResult<(Self, &str)> {
        if let Some(intersections) = unparsed.strip_prefix('&') {
            let (union, unparsed) = Self::parse_expr1(intersections)?;
            Self::parse_op2(Self::and(acc, union), unparsed)
        } else {
            Ok((acc, unparsed))
        }
    }

    /// Parses a complete expression
    ///
    /// It returns the unparsed text that follows
    fn parse_expr2(unparsed: &str) -> ColmenaResult<(Self, &str)> {
        let (union, unparsed) = Self::parse_expr1(unparsed)?;
        Self::parse_op2(union, unparsed)
    }

    /// Creates a new filter using an expression passed using `--on`.
    pub fn new<S: AsRef<str>>(filter: S) -> ColmenaResult<Self> {
        let filter = filter.as_ref();
        let trimmed = filter.trim();

        if trimmed.is_empty() {
            log::warn!("Filter \"{}\" is blank and will match nothing", filter);

            return Ok(Self::Empty);
        }
        let (target_filter, unparsed) = Self::parse_expr2(trimmed)?;
        if unparsed != "" {
            Err(ColmenaError::Unknown {
                message: format!("Found garbage {:?} when parsing the node filter.", unparsed),
            })
        } else {
            Ok(target_filter)
        }
    }

    /// Returns which NodeConfig information is needed to evaluate the filter.
    ///
    /// Evaluating `config.deployment` can potentially be very expensive,
    /// especially when its values (e.g., tags) depend on other parts of
    /// the configuration.
    pub fn needs_eval(&self) -> NeedsEval {
        // XXX: is the hashset overkill?
        match self {
            Self::MatchName(_) => NeedsEval {
                aliases: true,
                ..Default::default()
            },
            Self::MatchTag(_) => NeedsEval {
                tags: true,
                ..Default::default()
            },
            Self::Union(v) | Self::Inter(v) => v
                .iter()
                .fold(NeedsEval::default(), |acc, e| acc | e.needs_eval()),
            Self::Not(e) => e.needs_eval(),
            Self::Empty => NeedsEval::default(),
        }
    }

    /// Decides whether a node is accepted by the filter or not.
    /// panic if the filter depends on tags or aliases and they're None
    fn is_accepted(&self, name: &NodeName, config: &PartialNodeConfig) -> bool {
        match self {
            Self::MatchName(pat) => {
                pat.matches(name.as_str())
                    || config
                        .aliases
                        .as_ref()
                        .expect("aliases missing")
                        .iter()
                        .any(|alias| pat.matches(&alias.0))
            }
            Self::MatchTag(pat) => config
                .tags
                .as_ref()
                .expect("tags missing")
                .iter()
                .any(|tag| pat.matches(tag.as_str())),
            Self::Union(v) => v.iter().any(|e| e.is_accepted(name, config)),
            Self::Inter(v) => v.iter().all(|e| e.is_accepted(name, config)),
            Self::Not(e) => !e.is_accepted(name, config),
            Self::Empty => false,
        }
    }

    /// Runs the filter against a set of NodeConfigs and returns the matched ones.
    pub fn filter_nodes<'a, I>(&self, nodes: I) -> HashSet<NodeName>
    where
        I: Iterator<Item = (&'a NodeName, &'a PartialNodeConfig)>,
    {
        if self == &Self::Empty {
            return HashSet::new();
        }

        nodes
            .filter_map(|(name, config)| {
                if self.is_accepted(name, config) {
                    Some(name)
                } else {
                    None
                }
            })
            .cloned()
            .collect()
    }

    /// In case of trivial filters which dont actually use any node info
    /// Try to eval them immediately
    pub fn try_eval_trivial(&self) -> Option<bool> {
        match self {
            Self::MatchName(_) => None,
            Self::MatchTag(_) => None,
            Self::Union(fs) => {
                for f in fs {
                    match f.try_eval_trivial() {
                        None => return None,
                        Some(true) => return Some(true),
                        Some(false) => continue,
                    }
                }
                Some(false)
            }
            Self::Inter(fs) => {
                for f in fs {
                    match f.try_eval_trivial() {
                        None => return None,
                        Some(true) => continue,
                        Some(false) => return Some(false),
                    }
                }
                Some(true)
            }
            Self::Not(f) => f.try_eval_trivial().map(|b| !b),
            Self::Empty => Some(true),
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    impl PartialNodeConfig {
        fn known_empty() -> Self {
            Self {
                aliases: Some(vec![]),
                tags: Some(vec![]),
            }
        }

        pub fn known_aliases_tags(
            aliases: Option<Vec<NodeName>>,
            tags: Option<Vec<String>>,
        ) -> Self {
            Self { aliases, tags }
        }

        fn known_tags(tags: Vec<String>) -> Self {
            Self {
                aliases: Some(vec![]),
                tags: Some(tags),
            }
        }

        fn known_aliases(aliases: Vec<NodeName>) -> Self {
            Self {
                aliases: Some(aliases),
                tags: Some(vec![]),
            }
        }
    }

    use std::collections::{HashMap, HashSet};

    macro_rules! node {
        ($n:expr) => {
            NodeName::new($n.to_string()).unwrap()
        };
    }

    #[test]
    fn test_empty_filter() {
        let filter = NodeFilter::new("").unwrap();
        assert_eq!(NodeFilter::Empty, filter);

        let filter = NodeFilter::new("\t").unwrap();
        assert_eq!(NodeFilter::Empty, filter);

        let filter = NodeFilter::new("    ").unwrap();
        assert_eq!(NodeFilter::Empty, filter);
    }

    #[test]
    fn test_empty_filter_rule() {
        assert!(NodeFilter::new(",").is_err());
        assert!(NodeFilter::new("a,,b").is_err());
        assert!(NodeFilter::new("a,b,c,").is_err());
    }

    #[test]
    fn test_filter_rule_mixed() {
        let filter = NodeFilter::new("@router,gamma-*").unwrap();
        assert_eq!(
            NodeFilter::Union(vec![
                Box::new(NodeFilter::MatchTag(GlobPattern::new("router").unwrap())),
                Box::new(NodeFilter::MatchName(GlobPattern::new("gamma-*").unwrap())),
            ]),
            filter,
        );

        let filter = NodeFilter::new("a, \t@b ,    c-*").unwrap();
        assert_eq!(
            NodeFilter::Union(vec![
                Box::new(NodeFilter::MatchName(GlobPattern::new("a").unwrap())),
                Box::new(NodeFilter::MatchTag(GlobPattern::new("b").unwrap())),
                Box::new(NodeFilter::MatchName(GlobPattern::new("c-*").unwrap())),
            ]),
            filter,
        );

        let filter = NodeFilter::new("a  & \t@b ,    c-*").unwrap();
        assert_eq!(
            NodeFilter::Inter(vec![
                Box::new(NodeFilter::MatchName(GlobPattern::new("a").unwrap())),
                Box::new(NodeFilter::Union(vec![
                    Box::new(NodeFilter::MatchTag(GlobPattern::new("b").unwrap())),
                    Box::new(NodeFilter::MatchName(GlobPattern::new("c-*").unwrap())),
                ])),
            ]),
            filter,
        );

        let filter = NodeFilter::new("( a  & \t@b ) ,  c-*").unwrap();
        assert_eq!(
            NodeFilter::Union(vec![
                Box::new(NodeFilter::Inter(vec![
                    Box::new(NodeFilter::MatchName(GlobPattern::new("a").unwrap())),
                    Box::new(NodeFilter::MatchTag(GlobPattern::new("b").unwrap())),
                ])),
                Box::new(NodeFilter::MatchName(GlobPattern::new("c-*").unwrap())),
            ]),
            filter,
        );

        let filter = NodeFilter::new("( a  & \t@b ) ,  !  c-*").unwrap();
        assert_eq!(
            NodeFilter::Union(vec![
                Box::new(NodeFilter::Inter(vec![
                    Box::new(NodeFilter::MatchName(GlobPattern::new("a").unwrap())),
                    Box::new(NodeFilter::MatchTag(GlobPattern::new("b").unwrap())),
                ])),
                Box::new(NodeFilter::Not(Box::new(NodeFilter::MatchName(
                    GlobPattern::new("c-*").unwrap()
                )))),
            ]),
            filter,
        );

        let filter = NodeFilter::new("( a  & \t@b ) ,  !!!  c-*").unwrap();
        assert_eq!(
            NodeFilter::Union(vec![
                Box::new(NodeFilter::Inter(vec![
                    Box::new(NodeFilter::MatchName(GlobPattern::new("a").unwrap())),
                    Box::new(NodeFilter::MatchTag(GlobPattern::new("b").unwrap())),
                ])),
                Box::new(NodeFilter::Not(Box::new(NodeFilter::MatchName(
                    GlobPattern::new("c-*").unwrap()
                )))),
            ]),
            filter,
        );
    }

    #[test]
    fn test_filter_nodes_names_only() {
        let nodes = vec![
            (node!("lax-alpha"), PartialNodeConfig::known_empty()),
            (node!("lax-beta"), PartialNodeConfig::known_empty()),
            (node!("sfo-gamma"), PartialNodeConfig::known_empty()),
        ];

        assert_eq!(
            &HashSet::from_iter([node!("lax-alpha")]),
            &NodeFilter::new("lax-alpha")
                .unwrap()
                .filter_nodes(nodes.iter().map(|x| (&x.0, &x.1))),
        );

        assert_eq!(
            &HashSet::from_iter([node!("lax-alpha"), node!("lax-beta")]),
            &NodeFilter::new("lax-*")
                .unwrap()
                .filter_nodes(nodes.iter().map(|x| (&x.0, &x.1))),
        );
    }

    #[test]
    fn test_filter_nodes() {
        let nodes: HashMap<NodeName, PartialNodeConfig> = HashMap::from([
            (
                node!("alpha"),
                PartialNodeConfig::known_tags(vec!["web".to_string(), "infra-lax".to_string()]),
            ),
            (
                node!("beta"),
                PartialNodeConfig::known_tags(vec!["router".to_string(), "infra-sfo".to_string()]),
            ),
            (
                node!("gamma-a"),
                PartialNodeConfig::known_tags(vec!["controller".to_string()]),
            ),
            (
                node!("gamma-b"),
                PartialNodeConfig::known_tags(vec!["ewaste".to_string()]),
            ),
            (
                node!("aliases-test"),
                PartialNodeConfig::known_aliases_tags(
                    Some(vec![node!("whatever-alias1"), node!("whatever-alias2")]),
                    Some(vec!["testing".into()]),
                ),
            ),
        ]);
        assert_eq!(5, nodes.len());

        assert_eq!(
            &HashSet::from_iter([node!("alpha")]),
            &NodeFilter::new("@web").unwrap().filter_nodes(nodes.iter()),
        );

        assert_eq!(
            &HashSet::from_iter([node!("alpha"), node!("beta")]),
            &NodeFilter::new("@infra-*")
                .unwrap()
                .filter_nodes(nodes.iter()),
        );

        assert_eq!(
            &HashSet::from_iter([node!("beta"), node!("gamma-a")]),
            &NodeFilter::new("@router,@controller")
                .unwrap()
                .filter_nodes(nodes.iter()),
        );

        assert_eq!(
            &HashSet::from_iter([node!("beta"), node!("gamma-a"), node!("gamma-b")]),
            &NodeFilter::new("@router,gamma-*")
                .unwrap()
                .filter_nodes(nodes.iter()),
        );

        assert_eq!(
            &HashSet::from_iter([]),
            &NodeFilter::new("@router&@controller")
                .unwrap()
                .filter_nodes(nodes.iter()),
        );

        assert_eq!(
            &HashSet::from_iter([node!("beta")]),
            &NodeFilter::new("@router&@infra-*")
                .unwrap()
                .filter_nodes(nodes.iter()),
        );

        assert_eq!(
            &HashSet::from_iter([node!("alpha")]),
            &NodeFilter::new("!@router&@infra-*")
                .unwrap()
                .filter_nodes(nodes.iter()),
        );

        assert_eq!(
            &HashSet::from_iter([node!("aliases-test")]),
            &NodeFilter::new("whatever-alias1")
                .unwrap()
                .filter_nodes(nodes.iter()),
        );
    }
}
